from __future__ import annotations
from typing import TYPE_CHECKING, List, Tuple, Dict, Union, Optional, Any
if TYPE_CHECKING:
    pass

import re
import copy
import collections
from enum import Enum
from pathlib import Path

import numpy as np


class Mol2Headers(Enum):
    ALT_TYPE = "@<TRIPOS>ALT_TYPE"
    ANCHOR_ATOM = "@<TRIPOS>ANCHOR_ATOM"
    ASSOCIATED_ANNOTATION = "@<TRIPOS>ASSOCIATED_ANNOTATION"
    ATOM = "@<TRIPOS>ATOM"
    BOND = "@<TRIPOS>BOND"
    CENTER_OF_MASS = "@<TRIPOS>CENTER_OF_MASS"
    CENTROID = "@<TRIPOS>CENTROID"
    COMMENT = "@<TRIPOS>COMMENT"
    CRYSIN = "@<TRIPOS>CRYSIN"
    DICT = "@<TRIPOS>DICT"
    DATA_FILE = "@<TRIPOS>DATA_FILE"
    EXTENSION_POINT = "@<TRIPOS>EXTENSION_POINT"
    FF_PBC = "@<TRIPOS>FF_PBC"
    FFCON_ANGLE = "@<TRIPOS>FFCON_ANGLE"
    FFCON_DIST = "@<TRIPOS>FFCON_DIST"
    FFCON_MULTI = "@<TRIPOS>FFCON_MULTI"
    FFCON_RANGE = "@<TRIPOS>FFCON_RANGE"
    FFCON_TORSION = "@<TRIPOS>FFCON_TORSION"
    LINE = "@<TRIPOS>LINE"
    LSPLANE = "@<TRIPOS>LSPLANE"
    MOLECULE = "@<TRIPOS>MOLECULE"
    NORMAL = "@<TRIPOS>NORMAL"
    QSAR_ALIGN_RULE = "@<TRIPOS>QSAR_ALIGN_RULE"
    RING_CLOSURE = "@<TRIPOS>RING_CLOSURE"
    ROTATABLE_BOND = "@<TRIPOS>ROTATABLE_BOND"
    SEARCH_DIST = "@<TRIPOS>SEARCH_DIST"
    SEARCH_OPTIONS = "@<TRIPOS>SEARCH_OPTIONS"
    SET = "@<TRIPOS>SET"
    SUBSTRUCTURE = "@<TRIPOS>SUBSTRUCTURE"
    U_FEAT = "@<TRIPOS>U_FEAT"
    UNITY_ATOM_ATTR = "@<TRIPOS>UNITY_ATOM_ATTR"
    UNITY_BOND_ATTR = "@<TRIPOS>UNITY_BOND_ATTR"


MOL2_HEADER_INDICATOR = "@<TRIPOS>"
MOL2_HEADER_STARTING_MOL2_BLOCK = Mol2Headers.MOLECULE.value


def find_nth_instance_of_line_starting_with_substring(lines: List[str], substring: str, n: int) -> Optional[int]:
    """
    Finds the index of the nth line that starts with the given substring.

    Parameters:
        lines (List[str]): List of lines to search.
        substring (str): Substring to find.
        n (int): The nth instance of the substring to find.

    Returns:
        Optional[int]: The index of the nth line that starts with the substring, or None if not found.
    """

    if n < 1:
        raise ValueError("n must be >= 1")

    num_found = 0
    for i, line in enumerate(lines):
        if line.startswith(substring):
            num_found += 1
            if num_found == n:
                return i
    return None


def remove_leading_invalid_mol2_lines(mol2_lines: List[str]) -> List[str]:
    """Removes leading lines that are invalid for a mol2 block from a list of mol2 lines."""

    new_lines = copy.copy(mol2_lines)
    while len(new_lines) > 0:
        first_line = new_lines[0].strip()
        if first_line.startswith(MOL2_HEADER_INDICATOR) or first_line.startswith(
                "#"):  # only valid start to a mol2 block
            break
        new_lines.pop(0)

    return new_lines


def get_leading_comment_block_end_index(lines: List[str]) -> Union[int, None]:
    """Get the end index of the leading comment block if it exists."""

    leading_comment_block_end_index = None
    found_leading_comment_block = False
    for i in range(len(lines)):
        line = lines[i].strip()
        if line.startswith("#"):
            found_leading_comment_block = True
            leading_comment_block_end_index = i
        elif line == "":
            if found_leading_comment_block:
                leading_comment_block_end_index = i
            continue
        elif line != "":
            break

    return leading_comment_block_end_index


def get_trailing_comment_block_start_index(lines: List[str]) -> Union[int, None]:
    """Get the start index of the trailing comment block if it exists."""

    trailing_comment_block_start_index = None
    for i in reversed(range(len(lines))):
        line = lines[i].strip()
        if line.startswith("#"):
            trailing_comment_block_start_index = i
        elif line == "":
            continue
        elif line != "":
            break

    return trailing_comment_block_start_index


def extract_leading_comment_block(lines: List[str]) -> List[str]:
    """Extracts leading comment lines from a list of lines."""

    leading_comment_end = get_leading_comment_block_end_index(lines)
    if leading_comment_end is not None:
        leading_comment_lines = lines[:leading_comment_end + 1]
    else:
        leading_comment_lines = []

    return leading_comment_lines


def remove_leading_comment_block(lines: List[str]) -> List[str]:
    """Removes the leading comment lines from a list of lines."""

    leading_comment_end = get_leading_comment_block_end_index(lines)
    if leading_comment_end is not None:
        del lines[:leading_comment_end + 1]

    return lines


def extract_trailing_comment_block(lines: List[str]) -> List[str]:
    """Extracts trailing comment lines from a list of lines."""

    trailing_comment_start = get_trailing_comment_block_start_index(lines)
    if trailing_comment_start is not None:
        trailing_comment_lines = lines[trailing_comment_start:]
    else:
        trailing_comment_lines = []

    return trailing_comment_lines


def remove_trailing_comment_block(lines: List[str]) -> List[str]:
    """Removes the trailing comment lines from a list of lines."""

    trailing_comment_start = get_trailing_comment_block_start_index(lines)
    if trailing_comment_start is not None:
        del lines[trailing_comment_start:]

    return lines


def get_text_block(
    rows: List[List[Any]],
    header: Optional[str] = None,
    column_alignment: str = "none",
    num_spaces_between_columns: int = 1,
    num_spaces_before_line: int = 0,
):
    """
    Generates a formatted text block from a list of rows.

    Parameters:
    -----------
    rows : list of list of any
        A list of rows, each of which is a list of items to be converted to strings.

    header : str, optional (default=None)
        An optional header text to be included at the top of the text block.

    column_alignment : str, optional (default="none")
        The text alignment for each column. Options are "left", "right", and "none".

    num_spaces_between_columns : int, optional (default=1)
        The number of spaces to insert between each column.

    num_spaces_before_line : int, optional (default=0)
        The number of spaces to insert before each line of text.

    Returns:
    --------
    str
        A formatted text block as a single string.

    Examples:
    ---------
    >>> get_text_block([[1, 20, 300], [4000, 50, 6]], header="Header", column_alignment="right")
    'Header
       1  20 300
    4000  50   6
    '
    """

    #
    if column_alignment not in ["left", "right", "none"]:
        raise ValueError(
            f"Invalid column_alignment value '{column_alignment}'. Valid options are 'left', 'right', and 'none'."
        )

    #
    formatted_lines = []
    if len(rows) > 0:
        rows = [[str(token) for token in row] for row in rows]

        max_row_size = max([len(row) for row in rows])
        columns = [
            [row[i] if i < len(row) else "" for row in rows] for i in range(max_row_size)
        ]
        column_max_token_length_list = [
            max([len(token) for token in column]) for column in columns
        ]
        spacing_between_columns = " " * num_spaces_between_columns
        formatted_lines = []
        for row in rows:
            formatted_tokens = []
            for i, token in enumerate(row):
                if column_alignment == "left":
                    formatted_token = token.ljust(column_max_token_length_list[i])
                elif column_alignment == "right":
                    formatted_token = token.rjust(column_max_token_length_list[i])
                elif column_alignment == "none":
                    formatted_token = token
                else:
                    raise ValueError(
                        f"Invalid column_alignment value '{column_alignment}'. Valid options are 'left', 'right', and 'none'."
                    )
                formatted_tokens.append(formatted_token)
            spacing_before_line = num_spaces_before_line * " "
            formatted_line = spacing_before_line + spacing_between_columns.join(
                formatted_tokens
            )
            formatted_lines.append(formatted_line)

    text_block = ""
    if header:
        text_block += f"{header}\n"
    text_block += "\n".join(formatted_lines)

    return text_block


class Mol2DataRecord(object):
    def __init__(self, record_lines: List[str]):
        header, data_rows = self.parse_record_lines(record_lines)
        self.header = header
        self.data_rows = data_rows

    def __str__(self):
        return get_text_block(
            rows=self.data_rows,
            header=self.header,
            column_alignment='left',
            num_spaces_between_columns=4,
            num_spaces_before_line=4,
        )

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def parse_record_lines(record_lines: List[str]) -> Tuple[str, List[List[str]]]:
        header = record_lines[0].strip()
        data_rows = [line.split() for line in record_lines[1:]]
        return header, data_rows

    @staticmethod
    def line_is_valid_starting_line_to_mol2_data_record(line: str) -> bool:
        """Returns True if the line is a valid starting line to a mol2 data record."""
        return line.strip().startswith(MOL2_HEADER_INDICATOR) or line.strip().startswith("#")


class Mol2Block:
    def __init__(self, mol2_block_lines: List[str]) -> None:
        leading_comment_block_end_index = get_leading_comment_block_end_index(mol2_block_lines)
        if leading_comment_block_end_index is None:
            self.comment_lines = []
            remaining_lines = mol2_block_lines
        else:
            if len(mol2_block_lines) <= leading_comment_block_end_index + 1:
                raise ValueError(f"No records found in the supplied mol2 block lines: {mol2_block_lines}")
            self.comment_lines = mol2_block_lines[:leading_comment_block_end_index + 1]
            remaining_lines = mol2_block_lines[leading_comment_block_end_index + 1:]

        # Validate the records
        mol2_headers = [h.value for h in Mol2Headers]
        for header in mol2_headers:
            count = len([line for line in remaining_lines if line.strip() == header])
            if count > 1:
                raise ValueError(f"More than one data record of type {header} found in the supplied mol2 block lines: {remaining_lines}")

        # Parse the records
        records_dict = collections.OrderedDict()
        for record in self.split_mol2_block_into_data_records(remaining_lines):
            record_type_name = record.header.replace(MOL2_HEADER_INDICATOR, '').strip()  # e.g., "@<TRIPOS>ATOM" -> "ATOM"
            records_dict[record_type_name] = record
        self.data_records = records_dict

    def __str__(self):
        return "\n".join(self.comment_lines + [str(record) for record in self.data_records.values()])

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def split_mol2_block_into_data_records(mol2_block_lines: List[str]) -> List[Mol2DataRecord]:
        """Splits the supplied mol2 block into its component data records."""

        if not Mol2DataRecord.line_is_valid_starting_line_to_mol2_data_record(mol2_block_lines[0]):
            raise ValueError("The first line of the supplied mol2 block is not a valid record header or comment.")

        records = []
        n = 1
        should_break = False
        while True:
            start = find_nth_instance_of_line_starting_with_substring(mol2_block_lines, MOL2_HEADER_INDICATOR, n)
            end = find_nth_instance_of_line_starting_with_substring(mol2_block_lines, MOL2_HEADER_INDICATOR, n + 1)
            if end is None:
                end = len(mol2_block_lines)
                should_break = True

            record_lines = extract_trailing_comment_block(mol2_block_lines[:start]) + remove_trailing_comment_block(
                mol2_block_lines[start:end])
            records.append(Mol2DataRecord(record_lines))

            if should_break:
                break
            n += 1
        return records


class Mol2File(Path):
    def __init__(self, path: str):
        super().__init__(path)

        self.blocks = self.read_mol2_blocks(self.resolve())

    @staticmethod
    def read_mol2_blocks(mol2_file_path) -> List[Mol2Block]:
        """
        Reads and parses Mol2 blocks from a file specified by the `self.path` attribute.

        This function assumes that the file at `self.path` is a Mol2 file and processes it to
        extract information about molecules, atoms, and bonds. It captures this information into
        a list of Mol2Block objects, each representing a molecule and its associated details.

        Returns:
        --------
        list of Mol2Block
            A list of Mol2Block objects.

        Assumptions:
        ------------
        - The file is well-formatted according to the Mol2 standard.
        - The Mol2Block class is available and properly defined.

        Example Usage:
        --------------
        ```
        mol2_file = Mol2File(path="some_file.mol2")
        mol2_blocks = mol2_file.read_mol2_blocks()
        ```

        Notes:
        ------
        This function relies on `self.path` attribute to determine the file to read.
        """

        with open(mol2_file_path, "r") as f:
            lines = [line.strip() for line in f.readlines()]

        mol2_blocks = Mol2File.split_mol2_file_lines_into_mol2_blocks(lines)

        return mol2_blocks

    @staticmethod
    def split_mol2_file_lines_into_mol2_blocks(mol2_lines: List[str]) -> List[Mol2Block]:
        """Split the supplied mol2 file lines into mol2 blocks."""

        # Remove leading invalid lines
        mol2_lines = remove_leading_invalid_mol2_lines(mol2_lines)

        #
        if not Mol2DataRecord.line_is_valid_starting_line_to_mol2_data_record(mol2_lines[0]):
            raise ValueError("The first line of the supplied mol2 block is not a valid record header or comment.")

        #
        mol2_blocks = []
        n = 1
        should_break = False
        while True:
            start = find_nth_instance_of_line_starting_with_substring(mol2_lines, MOL2_HEADER_STARTING_MOL2_BLOCK, n)
            end = find_nth_instance_of_line_starting_with_substring(mol2_lines, MOL2_HEADER_STARTING_MOL2_BLOCK, n + 1)
            if end is None:
                end = len(mol2_lines)
                should_break = True

            mol2_block_lines = extract_trailing_comment_block(mol2_lines[:start]) + remove_trailing_comment_block(mol2_lines[start:end])
            mol2_blocks.append(Mol2Block(mol2_block_lines))

            if should_break:
                break

            n += 1

        return mol2_blocks

    def write_mol2_file_with_molecules_cloned_and_transformed(
        self,
        rotation_matrix: np.ndarray,
        translation_vector: np.ndarray,
        write_path: str,
        num_applications: int = 1,
        bidirectional: bool = False,
    ) -> None:
        """
        Writes a new Mol2 file with molecules that are cloned and transformed based on
        given rotation matrix and translation vector parameters.

        Parameters:
        -----------
        rotation_matrix : numpy.ndarray
            A 3x3 rotation matrix to transform the molecule's coordinates.

        translation_vector : numpy.ndarray
            A 1x3 translation vector to transform the molecule's coordinates.

        write_path : str
            The path where the new Mol2 file will be written.

        num_applications : int, optional (default=1)
            The number of times the transformation should be applied to each molecule.

        bidirectional : bool, optional (default=False)
            If True, applies the transformation in both directions (forward and inverse).

        Examples:
        ---------
        # Given that `self` is an instance of a class that contains this method
        ```
        rotation_matrix = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        translation_vector = np.array([1, 1, 1])
        self.write_mol2_file_with_molecules_cloned_and_transformed(
            rotation_matrix,
            translation_vector,
            'output.mol2',
            num_applications=2,
            bidirectional=True
        )
        ```
        """

        # Validate rotation_matrix
        if not isinstance(rotation_matrix, np.ndarray) or rotation_matrix.shape != (3, 3):
            raise ValueError("`rotation_matrix` must be a 3x3 numpy.ndarray")

        # Validate translation_vector
        if not isinstance(translation_vector, np.ndarray) or translation_vector.shape != (3,):
            raise ValueError("`translation_vector` must be a 1x3 numpy.ndarray")

        # Validate write_path
        if not isinstance(write_path, str):
            raise ValueError("`write_path` must be a string")

        # Validate num_applications
        if not isinstance(num_applications, int) or num_applications < 1:
            raise ValueError("`num_applications` must be an integer greater than or equal to 1")

        # Validate bidirectional
        if not isinstance(bidirectional, bool):
            raise ValueError("`bidirectional` must be a boolean")

        #
        def transform(xyz: np.ndarray, rot_mat: np.ndarray, transl_vec: np.ndarray) -> np.ndarray:
            """Transforms the coordinates using the given rotation matrix and translation vector."""

            return np.dot(rot_mat, xyz) + transl_vec

        def get_inverse_transform(rot_mat: np.ndarray, transl_vec: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
            """Calculates the inverse transformation."""

            a = np.array([[1.0, 0.0, 0.0, 0.0]])
            for i in range(3):
                a = np.concatenate(
                    (
                        a,
                        np.array(
                            [np.concatenate((np.array([transl_vec[i]]), rot_mat[i, :]))]
                        ),
                    ),
                    axis=0,
                )
            a_inv = np.linalg.inv(a)
            rot_mat_inv = a_inv[1:, 1:]
            transl_vec_inv = a_inv[1:, 0]

            return rot_mat_inv, transl_vec_inv

        #
        if bidirectional:
            rotation_matrix_inv, translation_vector_inv = get_inverse_transform(
                rotation_matrix, translation_vector
            )

        #
        with open(write_path, "w") as f:
            new_mol2_blocks = []
            for mol2_block in copy.deepcopy(self.blocks):
                #
                new_molecule_record_data_rows = []
                new_molecule_record_data_rows.append(mol2_block.data_records[Mol2Headers.MOLECULE.name].data_rows[0])
                molecule_row = mol2_block.data_records[Mol2Headers.MOLECULE.name].data_rows[1]
                if bidirectional:
                    multiplier = (2 * num_applications) + 1
                else:
                    multiplier = num_applications + 1
                new_molecule_row = [
                    int(molecule_row[0]) * multiplier,
                    int(molecule_row[1]) * multiplier,
                ] + molecule_row[2:]
                new_molecule_record_data_rows.append(new_molecule_row)
                if len(mol2_block.data_records[Mol2Headers.MOLECULE.name].data_rows) > 2:
                    new_molecule_record_data_rows += mol2_block.data_records[Mol2Headers.MOLECULE.name].data_rows[2:]

                #
                atom_element_to_id_nums_dict = collections.defaultdict(list)
                atom_names = [atom_row[1] for atom_row in mol2_block.data_records[Mol2Headers.ATOM.name].data_rows]
                for atom_name in atom_names:
                    element, id_num = [
                        token for token in re.split(r"(\d+)", atom_name) if token
                    ]
                    atom_element_to_id_nums_dict[element].append(int(id_num))

                #
                num_atoms = len(mol2_block.data_records[Mol2Headers.ATOM.name].data_rows)
                new_atom_record_data_rows = []
                for atom_row in mol2_block.data_records[Mol2Headers.ATOM.name].data_rows:
                    new_atom_record_data_rows.append(atom_row)

                def apply_to_atoms(rot_mat: np.ndarray, transl_vec: np.ndarray, n: int, num_app: int) -> None:
                    """
                    Applies transformations to atoms and appends the transformed atoms to `new_atom_record_data_rows`.

                    Parameters:
                    -----------
                    rot_mat : np.ndarray
                        A 3x3 rotation matrix to apply to each atom's coordinates.

                    transl_vec : np.ndarray
                        A 1x3 translation vector to apply to each atom's coordinates.

                    n : int
                        An integer multiplier for generating new atom IDs and names.

                    num_app : int
                        The number of times the transformation should be applied to each atom.

                    Assumptions:
                    ------------
                    - The function assumes that `new_atom_record_data_rows` and `mol2_block`
                      are defined in the outer scope.
                    - The function also assumes that `num_atoms` and `atom_element_to_id_nums_dict` are
                      defined and appropriately populated in the outer scope.
                    """
                    if num_app < 1:
                        raise ValueError("num_app must be >= 1")

                    for atom_row in mol2_block.data_records[Mol2Headers.ATOM.name].data_rows:
                        atom_id = atom_row[0]
                        new_atom_id = f"{int(atom_id) + (n * num_atoms)}"
                        atom_name = atom_row[1]
                        element, id_num = [
                            token for token in re.split(r"(\d+)", atom_name) if token
                        ]
                        new_atom_name = f"{element}{int(id_num) + (n * max(atom_element_to_id_nums_dict[element]))}"
                        current_xyz = np.array(
                            [float(coord) for coord in atom_row[2:5]]
                        )
                        for j in range(num_app):
                            new_xyz = transform(current_xyz, rot_mat, transl_vec)
                            current_xyz = new_xyz
                        new_atom_row = (
                            [new_atom_id, new_atom_name] + list(new_xyz) + atom_row[5:]
                        )
                        new_atom_record_data_rows.append(new_atom_row)

                #
                for i, n in enumerate(list(range(1, num_applications + 1))):
                    apply_to_atoms(
                        rotation_matrix, translation_vector, n, num_app=i + 1
                    )

                #
                if bidirectional:
                    for i, n in enumerate(
                        list(range(num_applications + 1, (2 * num_applications) + 1))
                    ):
                        apply_to_atoms(
                            rotation_matrix_inv,
                            translation_vector_inv,
                            n,
                            num_app=i + 1,
                        )

                #
                num_bonds = len(mol2_block.data_records[Mol2Headers.BOND.name].data_rows)
                new_bond_record_data_rows = []
                for bond_row in mol2_block.data_records[Mol2Headers.BOND.name].data_rows:
                    new_bond_record_data_rows.append(bond_row)

                def apply_to_bonds(n):
                    """
                    Applies transformations to bonds and appends them to `new_bond_record_data_rows`.

                    This function takes an integer `n` and uses it to generate new bond IDs and associated atom IDs.
                    Transformed bond records are appended to the `new_bond_record_data_rows` list, assumed to be defined
                    in the outer scope.

                    Parameters:
                    -----------
                    n : int
                        An integer multiplier for generating new bond IDs and associated atom IDs.

                    Assumptions:
                    ------------
                    - Assumes `new_bond_record_data_rows` and `mol2_block` are defined in the outer scope.
                    - Assumes `num_atoms` and `num_bonds` are defined and appropriately populated in the outer scope.
                    """

                    for bond_row in mol2_block.data_records[Mol2Headers.BOND.name].data_rows:
                        new_bond_row = (
                            [int(bond_row[1]) + (n * num_bonds)]
                            + [int(num) + (n * num_atoms) for num in bond_row[1:3]]
                            + bond_row[3:]
                        )
                        new_bond_record_data_rows.append(new_bond_row)

                #
                for n in range(1, num_applications + 1):
                    apply_to_bonds(n)

                #
                if bidirectional:
                    for n in range(num_applications + 1, (2 * num_applications) + 1):
                        apply_to_bonds(n)

                #
                mol2_block.data_records[Mol2Headers.MOLECULE.name].data_rows = new_molecule_record_data_rows
                mol2_block.data_records[Mol2Headers.ATOM.name].data_rows = new_atom_record_data_rows
                mol2_block.data_records[Mol2Headers.BOND.name].data_rows = new_bond_record_data_rows

                #
                new_mol2_blocks.append(mol2_block)

            #
            f.write(f"{self.get_mol2_blocks_as_string(new_mol2_blocks)}\n")

    def __str__(self):
        """
        Returns string representation of the Mol2RecordSet object as would be written to a .mol2 file.
        """
        return Mol2File.get_mol2_blocks_as_string(self.blocks)

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def get_mol2_blocks_as_string(mol2_blocks: List[Mol2Block]) -> str:
        """
        Returns a string representation of a list of Mol2Block objects as would be written to a .mol2 file.
        """
        return "\n".join([str(mol2_block) for mol2_block in mol2_blocks])

